"""Configuration for session factory"""

from spynl.main.exceptions import SpynlException
from spynl.main.utils import get_parsed_body


def main(config):
    """
    * Configure which session factory to use, pyramid's default or one that
      is set in a plugin. In the plugger.py the plugin can set function by
      assigning that function to a setting with the name of the 'spynl.session'
    * Get session id from request to retrieve session
    * Save current session after each response
    """
    settings = config.get_settings()
    if 'spynl.session' in settings and settings['spynl.session'] != 'Pyramid':

        Session = settings.get(settings['spynl.session'])
        if not Session:
            msg = ('The function for the session factory is not set in any of '
                   'the plugins for this spynl.session option: {}'
                   .format(settings['spynl.session']))
            raise SpynlException(msg)

        def mksession(sid):
            return Session(id=sid)
    else:
        from beaker.session import Session

        def mksession(sid):
            return Session(None, id=sid, use_cookies=False)

    def session_factory(request):
        """We keep one session per sid"""
        sid = None
        # getting the sid from request.args would be easier,
        # but we cannot rely on args being unified there already.
        # We look for a session ID (sid) with decreasing priority in
        # cookies, headers, GET vars, POST vars.
        try:
            body = get_parsed_body(request)
        except:
            body = {}
        for params in (body, request.GET,
                       request.headers, request.cookies):
            if 'sid' in params:
                sid = params['sid']

        return mksession(sid)

    config.set_session_factory(session_factory)

    def new_response(event):
        """
        Make sure new and pristine sessions are saved.
        Changes to sessions should save them automatically
        and set session.new to False.
        """
        session = event.request.session
        if session:
            if hasattr(session, 'new'):
                if session.new is True:
                    session.save()
            else:
                session.save()

    config.add_subscriber(new_response, 'pyramid.events.NewResponse')



"""
Mercurial-specific tasks to use across repos.
"""
import os

from invoke import task
from invoke.exceptions import Exit
import pip

from spynl.main.pkg_utils import get_spynl_package
from cli.utils import resolve_repo_names, repo
from cli import dev


@task(help={'repos': dev.tasks.repos_help})
def pullup(ctx, repos='_all'):
    '''Pull from repo(s) and update, whatever branch they're on.'''
    for repo_name in resolve_repo_names(repos):
        print("[SPYNL] Pulling repo: {}".format(repo_name))
        with repo(repo_name):
            ctx.run('hg pull --update')


@task(help={'repos': dev.tasks.repos_help,
            'show-all': 'If True, show all inactive branches as well. '
                        'Defaults to False.'})
def branches(ctx, repos='_all', show_all=False):
    '''Show the active branch in repos.'''
    for repo_name in resolve_repo_names(repos):
        with repo(repo_name):
            print('[SPYNL] Branches for {}:'.format(repo_name))
            if show_all:
                ctx.run('hg branches -q')
            else:
                ctx.run('hg branch -q')


@task(help={'repos': dev.tasks.repos_help,
            'revision ': 'The revision (e.g. branch or commit ID) '
                         'to update to'})
def update(ctx, repos='_all', revision='default'):
    '''Update all repos to a specific revision.'''
    for repo_name in resolve_repo_names(repos):
        print('[SPYNL] Updating repo {}:'.format(repo_name))
        with repo(repo_name):
            ctx.run('hg update {}'.format(revision), warn=True)


@task(help={'repos': dev.tasks.repos_help})
def status(ctx, repos='_all'):
    '''Show status in repos.'''
    for repo_name in resolve_repo_names(repos):
        print('[SPYNL] Status for {}:'.format(repo_name))
        with repo(repo_name):
            ctx.run('hg status')


@task(help={'repos': dev.tasks.repos_help})
def outgoing(ctx, repos='_all'):
    '''Show if there are outgoing (to be pushed) commits.'''
    for repo_name in resolve_repo_names(repos):
        with repo(repo_name):
            print('')
            print('[SPYNL] Outgoing commits for {}:'.format(repo_name))
            ctx.run('hg outgoing', warn=True)


@task(help={'repos': dev.tasks.repos_help})
def diff(ctx, repos='_all'):
    '''Show diff in repos.'''
    for repo_name in resolve_repo_names(repos):
        with repo(repo_name):
            ctx.run('hg diff')


@task(help={'repos': dev.tasks.repos_help,
            'tagname': 'Name of the tag. Required.',
            'force': 'Add the -f option to the mercurial tag command. '
                     'Defaults to True.'})
def tag(ctx, repos='_all', tagname=None, force=True):
    '''Tag current state in repos.'''
    if tagname is None:
        raise Exit("Error: No tagname argument given.")
    for repo_name in resolve_repo_names(repos):
        with repo(repo_name):
            if force:
                ctx.run('hg tag %s -f' % tagname)
            else:
                ctx.run('hg tag %s' % tagname)


@task(help={'repos': dev.tasks.repos_help,
            'msg': 'Commit Message'})
def commit(ctx, repos='_all', msg=''):
    '''Commit changed code.'''
    for repo_name in resolve_repo_names(repos):
        print("[SPYNL] Commiting repo: {}".format(repo_name))
        # mercurial returns non-zero error code when no changes
        with repo(repo_name):
            if msg != "":
                ctx.run("hg commit -m '{}'".format(msg), warn=True)
            else:
                ctx.run("hg commit", warn=True)


@task(optional=['skip-testing'],
      help={'repos': dev.tasks.repos_help,
            'skip-testing': 'Tests are skipped when this param is given.'})
def push(ctx, repos='_all', skip_testing=False):
    '''Push commited code to server, run tests first.'''
    if not skip_testing:
        for repo_name in resolve_repo_names(repos):
            dev.tasks.test(ctx, repos=repo_name, called_standalone=False)
    for repo_name in resolve_repo_names(repos):
        print("[SPYNL] Pushing repo: {}".format(repo_name))
        with repo(repo_name):
            # mercurial returns non-zero error code when no changes
            ctx.run("hg push", warn=True)


@task(help={'repos': dev.tasks.repos_help,
            'sourcebranch': 'Branch to merge from. Required.',
            'targetbranch': 'Branch to merge into. Defaults to "default"'})
def merge(ctx, repos='_all', sourcebranch=None, targetbranch='default'):
    '''Merge a source branch into the target branch.'''
    if sourcebranch is None:
        print("[SPYNL] No target branch given.")
        return
    spynl = get_spynl_package('spynl')
    for repo_name in resolve_repo_names(repos):
        print("[SPYNL] Merging branch {} of repo: {} into branch {}"
              .format(sourcebranch, repo_name, targetbranch))
        with repo(repo_name):
            ctx.run('{}/cli/dev/merge_branch.sh {} {}'
                    .format(spynl.location, sourcebranch, targetbranch))


@task(help={'repos': dev.tasks.repos_help,
            'name': 'The new branch name.'})
def new_branch(ctx, repos='_all', name=None):
    ''' Create a new branch.'''
    if not name:
        print("[SPYNL] No branch name given.")
        return
    spynl = get_spynl_package('spynl')
    for repo_name in resolve_repo_names(repos):
        print("[SPYNL] Making new branch named {} of repo: {}"
              .format(name, repo_name))
        with repo(repo_name):
            ctx.run('%s/cli/dev/new_branch.sh %s' % (spynl.location, name))



"""Helper functions for tests to use."""


import json


def post(application, view, params=None, content_type='application/json',
         expect_errors=False, return_headers=False):
    """
    Make a POST request and get a json object back.

    Helper method to save typing.
    """
    if params is None:
        params = {}
    if 'json' in content_type:
        paramstr = json.dumps(params)
    response = application.post(view, params=paramstr,
                                headers={'Content-Type': content_type},
                                expect_errors=expect_errors)
    if response.text:
        rtext = json.loads(response.text)
        if return_headers:
            return rtext, response.headers
        else:
            return rtext
    if return_headers:
        return response.headers


def get(application, path, expect_errors=False, return_headers=False,
        headers=None):
    """
    Make a GET request and get a json object back.

    Helper method to save typing.
    """
    response = application.get(path, expect_errors=expect_errors,
                               headers=headers)
    if response.text:
        rtext = json.loads(response.text)
        if return_headers:
            return rtext, response.headers
        else:
            return rtext
    if return_headers:
        return response.headers



"""Define functions to send emails."""


import os

from pyramid.renderers import render
from pyramid_mailer import get_mailer
from pyramid_mailer.mailer import DummyMailer
from pyramid_mailer.message import Message

from jinja2 import Environment, Template, TemplateNotFound, FileSystemLoader
import html2text

from spynl.main.utils import (get_logger, get_settings,
                              is_production_environment)
from spynl.main.exceptions import EmailTemplateNotFound
from .locale import TemplateTranslations, SpynlTranslationString


DEFAULT_HTML_TEMPLATE = '''
    <!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"
                        "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
    <html xmlns="http://www.w3.org/1999/xhtml">
      <head>
        <meta http-equiv="Content-Type" content="text/html; charset=UTF-8" />
        <title>{{subject}}</title>
        <meta name="viewport" content="width=device-width, initial-scale=1.0"/>
      </head>
      <body style="margin: 0; padding: 0;">
        <!-- Spynl Default Email Template -->
        <table align="center" cellpadding="0" cellspacing="0" width="400">
        <tr>
          <td align="center" bgcolor="#fff" style="padding: 40px 0 30px 0;"
              style="color: #153643; font-family: Arial, sans-serif;
                     font-size: 16px; line-height: 20px;">
            {{content}}
          </td>
        </tr>
        </table>
      </body>
    </html>'''


def _sendmail(request, recipient, subject, plain_body, html_body=None,
              sender=None, attachments=None, fail_silently=False, mailer=None):
    """
    Send a mail using the pyramid mailer. This function also makes sure
    that if Spynl is not in a production environment, mail is sent to a dummy
    email address.

    :param Request request: the original request
    :param string recipient: addressee
    :param string subject: subject of mail
    :param string plain_body: content of mail
    :param string html_body: html content of mail
    :param string sender: senders address (optional,
                          default: no_reply@<spynl.domain>)
    :param [Attachment] attachments: attachments (optional)
    :param bool fail_silently: keep quiet when connection errors happen
    :param pyramid_mailer.Mailer mailer: mailer object (optional)
    """
    settings = get_settings()
    logger = get_logger()

    if mailer is None:
        mailer = get_mailer(request)

    if not sender:
        sender = settings.get('mail.sender')
        if not sender:
            domain = settings.get('spynl.domain', 'example.com')
            sender = 'no_reply@{}'.format(domain)

    sender = str(sender)
    recipients = None
    # use recipient in unit tests (mailer only writes files) or in real env
    if recipient and (mailer.__class__ is DummyMailer or
                      is_production_environment()):
        recipients = [str(recipient)]
    # otherwise (e.g. during dev) we might use dummy mail address
    elif settings.get('mail.dummy_recipient'):
        recipients = [settings.get('mail.dummy_recipient')]
        logger.info("I will send this email to %s instead of %s.",
                    settings.get('mail.dummy_recipient'), recipient)
    subject = str(subject).rstrip()

    message = Message(sender=sender, recipients=recipients, subject=subject,
                      body=plain_body, html=html_body)
    if attachments:
        for attachment in attachments:
            message.attach(attachment)

    try:
        message.validate()
        logger.info('Sending email titled "%s" to %s from %s',
                    subject, recipients, sender)
        mailer.send_immediately(message, fail_silently=fail_silently)
    except Exception as e:
        logger.exception(e)
        return False

    return True


def send_template_email(request, recipient, template_string=None,
                        template_file=None, plugin=None, replacements={},
                        subject='', fail_silently=False, mailer=None,
                        attachments=None, sender=None):
    """
    Send email using a html template.
    
    The email's content (which can contain HTML code) is defined by a Jinja
    template. This content template can be given either by filename
    <template_file> (absolute path) or directly in string form
    <template_string>. In both cases, the email content gets wrapped by a base
    template which defines a consistent layout. The base template to use is
    defined by a string type setting:
        `base_email_template: absolute path of template
    If the "base template" cannot be loaded the <DEFAULT_HTML_TEMPLATE> is
    used.
    Replacements can be send in to customise the email content.
    If no html or plain text was constructed in the end, replacements are being
    used to construct a basic text to be send as an email.

    In case of a jinja2 template the following assignments are attempted to be
    loaded from the template:
        `default_subject`: will be used if no subject was passed
    """
    if (template_string is None and template_file is None) or \
            (template_string is not None and template_file is not None):
        raise Exception('One of <template_string> or <template_file> must be '
                        'given.')
    text_body = ''
    if subject:
        replacements['subject'] = subject
    for key, value in replacements.items():
        if isinstance(value, SpynlTranslationString):
            replacements[key] = value.translate()

    if template_file is not None:
        try:
            replacements['content'] = render(template_file, replacements,
                                             request=request)
        except TemplateNotFound:
            raise EmailTemplateNotFound(template_file)

        if not subject:  # update with possible subject defined in template
            subject = value_from_template(template_file, 'default_subject',
                                          request, replacements, plugin)
            replacements.update(subject=subject)
    else:
        replacements['content'] = Template(template_string).render(
            **replacements)

    # Render the base template with the content of the given template
    base_template = get_settings().get('base_email_template')
    if base_template is not None:
        html_body = render(base_template, replacements, request=request)
    else:
        html_body = Template(DEFAULT_HTML_TEMPLATE).render(**replacements)
    html_body = html_body.replace('\n', '')

    text_maker = html2text.HTML2Text()
    text_maker.ignore_images = True
    text_body = text_maker.handle(html_body)

    if not text_body and not html_body:
        str_replacements = '\n'.join(['{}: {}'.format(key, value)
                                      for key, value in replacements.items()])
        text_body = 'Your account has been changed.\n' + str_replacements
        template = Template(DEFAULT_HTML_TEMPLATE)
        replacements.update(subject=subject, content=text_body)
        html_body = template.render(**replacements)
        get_logger(__name__).error(
            'Body was not found in body_string or email template: %s',
            template)

    return _sendmail(request, recipient, subject, text_body, html_body,
                     fail_silently=fail_silently, mailer=mailer,
                     attachments=attachments, sender=sender)


def value_from_template(template_file, var, request, replacements,
                        plugin=None):
    """
    Return the value of given <var> defined inside the <template_file>.

    Set the static_url filter for templates to construct urls for static files.
    If plugin was given, install plugin's translations to the environment. 
    """
    path, filename = os.path.split(template_file)
    env = Environment(loader=FileSystemLoader(path),
                      extensions=['jinja2.ext.i18n'])
    env.filters['static_url'] = request.static_url
    if plugin is not None:
        translations = TemplateTranslations(plugin)
        env.install_gettext_translations(translations)

    template = env.get_template(filename)
    template_module = template.make_module(vars=replacements)
    return getattr(template_module, var, '')



import os
from invoke import task
from invoke.exceptions import Exit
import pip

from spynl.main.version import __version__ as spynl_version
from spynl.main.pkg_utils import  get_config_package, get_dev_config 
from cli.utils import resolve_repo_names, repo, chdir


repos_help = "Affected repos, e.g. spynl,spynl.auth. "\
             "Defaults to all installed."


@task(help={'repos': repos_help,
            'developing': 'If True (default), python setup.py develop is '
                          'used, otherwise install',
            'revision': 'Which revision to update to (default: do not update)',
            'fallbackrevision': 'Use this revision if the repository does not '
                                ' have the target revision.',
            'install-deps': 'If True (default), pre-install dependencies '
                            '(eg. via apt-get).',
            'src_path': 'Desired installation path. Defaults to '
                        '$VIRTUAL_ENV/src directory.'})
def install(ctx, repos='_all', developing=True, revision=None,
            fallbackrevision=None, install_deps=True, src_path=None):
    '''Get code and install repos in the current virtual environment.'''
    ctx.run('pip install --upgrade setuptools')  # make sure we have the latest
    src_path = src_path or os.environ['VIRTUAL_ENV'] + '/src'
    config_package = get_config_package()
    if config_package is None:
        raise Exit("No config package with ini files found.")
    config = get_dev_config() 
    if fallbackrevision is None:
        fallbackrevision = 'master' 
        if config.get('spynl.ops.scm_type', '') == 'hg':
            fallbackrevision = 'default' 
    if not os.path.exists(src_path):
        os.mkdir(src_path)
    for repo_name in resolve_repo_names(repos, complain_not_installed=False):
        repo_path = src_path + '/' + repo_name
        if not os.path.exists(repo_path):
            clone_url = 'ssh://%s@%s/%s' % (config.get('spynl.ops.scm_type'),
                                            config.get('spynl.ops.scm_team'),
                                            repo_name)
            with chdir(src_path):
                print("[spynl dev.install] CLONING from %s" % clone_url)
                ctx.run('hg clone %s' % clone_url)
        else:
            print("[spynl dev.install] UPDATING {}".format(repo_name))
            with chdir(repo_path):
                ctx.run('hg pull --update')
        with chdir(repo_path):
            if revision is not None:
                print('[spynl dev.install] Updating to revision %s ...' % revision)
                if not ctx.run('hg update {}'.format(revision), warn=True):
                    print('[spynl dev.install] Updating to fallback revision %s ...'
                          % fallbackrevision)
                    ctx.run('hg update {}'.format(fallbackrevision))
            if os.path.isfile('./setup.sh'):
                pre_cmd = './setup.sh --pre-install '\
                          '--virtualenv=%s' % os.environ['VIRTUAL_ENV']
                if install_deps:
                    pre_cmd += ' --install-dependencies'
                ctx.run(pre_cmd)
            if developing:
                print("[spynl dev.install] DEVELOPING {}".format(repo_name))
                ctx.run('python setup.py develop')
            else:
                print("[spynl dev.install] INSTALLING {}".format(repo_name))
                ctx.run('python setup.py install')
            translations(ctx, repos=repo_name, action='compile')
            if os.path.isfile('./setup.sh'):
                ctx.run('./setup.sh --post-install')


@task
def serve(ctx):
    '''Run local dev server'''
    config_package = get_config_package()
    if config_package is None:
        raise Exit("No spynl-plugin found with .ini files. Exiting ...")
    with repo('spynl'):
        ctx.run('pserve %s/development.ini --reload'
                % config_package.location, pty=True)


@task(aliases=('tests',),
      help={'repos': repos_help,
            'called-standalone': 'If False, this task is a pre-task and the '
                                 'user gets to cancel the task flow when '
                                 'tests fail.',
            'reports': 'If True, Junit and Coverage reports will be made '
                       '(requires a few extra packages, '
                       'e.g. for pycoverage).'})
def test(ctx, repos='_all', called_standalone=True, reports=False):
    '''
    Perfom tests in one or more spynl plugins.
    '''
    for repo_name in resolve_repo_names(repos):
        print("[spynl dev.test] Testing repo: {}".format(repo_name))
        with repo(repo_name):
            if not reports:
                result = ctx.run('py.test', warn=True)
            else:
                result = ctx.run('py.test --junit-xml=pytests.xml --cov %s '
                                 ' --cov-report xml --cov-append' % repo_name)
            if not result.ok and called_standalone is False\
               and input("[spynl dev.test] Tests failed. Continue anyway? Y/n") not in ('y', 'Y'):
                raise Exit("[spynl dev.test] Aborting at user request.")


@task(help={'repos': repos_help,
            'languages': 'An iterable of language codes. Defaults to ("nl",).',
            'action': 'Either "compile" (compile selected languages) or '
                      '"refresh" (extract messages from the source code '
                      'and update the .po files for the selected languages). '
                      'Defaults to compile.'})
def translations(ctx, repos='_all', languages=('nl',), action='compile'):
    '''
    Ensure that translations files are up-to-date w.r.t. to the code. If action
    is set to compile (default), this will compile the catalogs for all selected
    languages. If action is set to refresh, this will extract messages from the
    source code and update the .po files for the selected languages (will
    initialize if necessary).
    '''
    if action == 'compile':
        refresh = False
    elif action == 'refresh':
        refresh = True
    else:
        raise Exit("[spynl dev.translations] action should be 'compile' or "
                   "'refresh'.")

    for repo_name in resolve_repo_names(repos):
        with repo(repo_name):
            if repo_name == 'spynl':
                repo_name = 'spynl.main'
            print("[spynl dev.translations] Repo: %s ..." % repo_name)
            repo_path = repo_name.replace('.', '/')
            # make locale folder if it doesn't exist already:
            if not os.path.exists('%s/locale' % repo_path):
                os.mkdir('./%s/locale' % repo_path)
            if refresh:
                # Extract messages from source:
                ctx.run('python setup.py extract_messages '
                        '--output-file {rp}/locale/messages.pot --no-wrap '
                        '--sort-by-file --input-dirs {rp} --project {project} '
                        '--copyright-holder "Softwear BV" --version {v}'
                        .format(rp=repo_path, project=repo_name,
                                v=spynl_version))
            for lang in languages:
                path2po = '%s/locale/%s/LC_MESSAGES/%s.po' % (repo_path, lang,
                                                              repo_name)
                # update if refresh
                if refresh:
                    # init if needed:
                    if not os.path.exists(path2po):
                        print('[spynl dev.translations] File %s does not exist.'
                              ' Initializing.' % path2po)
                        ctx.run('python setup.py init_catalog -l {lang} '
                                '-i {rp}/locale/messages.pot '
                                '-d {rp}/locale -D {rn}'
                                .format(rp=repo_path, rn=repo_name, lang=lang))
                    # update if not init
                    else:
                        print('[spynl dev.translations] update the %s catalog'
                              % lang)
                        ctx.run('python setup.py update_catalog -N --no-wrap '
                                '-l {lang} -i {rp}/locale/messages.pot '
                                '-d {rp}/locale -D {rn}'
                                .format(rp=repo_path, rn=repo_name, lang=lang))
                # if not refresh, compile
                elif os.path.exists(path2po):
                    ctx.run('python setup.py compile_catalog --domain {rn} '
                            '--directory {rp}/locale --domain {rn} '
                            '--locale {lang}'
                            .format(rp=repo_path, rn=repo_name, lang=lang))
                else:
                    print('[spynl dev.translations] File %s does not exist.'
                          ' Update the repo first.' % path2po)

                print("[spynl dev.translations] Done with language %s." % lang)
                print("--------------------------------------------------")



"""Helper functions and view derivers for spynl.main."""


import json
import logging
import traceback
import sys
from functools import wraps
from inspect import isfunction, isclass, getargspec
from collections import namedtuple
import yaml
from tld import get_tld
from tld.exceptions import TldBadUrl, TldDomainNotFound
import inspect

from pyramid.response import Response
from pyramid.renderers import json_renderer_factory
from pyramid.exceptions import Forbidden
from pyramid import threadlocal

from spynl.main import urlson
from spynl.main.exceptions import SpynlException, MissingParameter, BadOrigin
from spynl.main.version import __version__ as spynl_version
from spynl.main.locale import SpynlTranslationString as _



def get_request():
    """
    Retrieve current request.

    Use with care, though:
    http://docs.pylonsproject.org/projects/pyramid/en/latest/narr/threadlocals.html
    """
    return threadlocal.get_current_request()


def get_settings():
    """
    Get settings (from .ini file [app:main] section)
    Can also be accessed from the request object: request.registry.settings
    For more info on the way we do it here, consult
    http://docs.pylonsproject.org/projects/pyramid/en/latest/narr/threadlocals.html
    Our policy is to not edit the settings during a request/response cycle.
    """
    registry = threadlocal.get_current_registry()
    return registry.settings if registry.settings is not None else {}


def check_origin(endpoint, info):
    """Check if origin is allowed"""
    def wrapper_view(context, request):
        """raise HTTPForbidden if origin isn't allowed"""
        origin = request.headers.get('Origin', '')
        if not is_origin_allowed(origin):
            # because this is a wrapper, the bad origin will not be properly
            # escalated to forbidden, so it needs to be done like this.
            raise Forbidden(detail=BadOrigin(origin).message.translate(
                request.localizer))
        return endpoint(context, request)
    return wrapper_view


def handle_pre_flight_request(endpoint, info):
    """
    "pre-flight-request": return custom response with some information on
    what we allow. Used by browsers before they send XMLHttpRequests.
    """
    def wrapper(context, request):
        """Call the endpoint if not an OPTION (pre-flight) request,
        otherwise return a custom Response."""
        if request.method != 'OPTIONS':
            return endpoint(context, request)
        else:
            headerlist = []
            origin = request.headers.get('Origin')
            if origin:  # otherwise we are on localhost or are called directly
                if is_origin_allowed(origin):
                    headerlist.append(('Access-Control-Allow-Origin', origin))
                else:
                    headerlist.append(('Access-Control-Allow-Origin', 'null'))
            headerlist.extend([('Access-Control-Allow-Methods', 'GET,POST'),
                               ('Access-Control-Max-Age', '86400'),
                               ('Access-Control-Allow-Credentials', 'true'),
                               ('Content-Length', '0'),
                               ('Content-Type', 'text/plain')])
            # you can send any headers to Spynl, basically
            if 'Access-Control-Request-Headers' in request.headers:
                headerlist.append(
                    ('Access-Control-Allow-Headers',
                     request.headers['Access-Control-Request-Headers']))
            # returning a generic and resource-agnostic pre-flight response
            return Response(headerlist=headerlist)
    return wrapper


def is_origin_allowed(origin):
    """
    Check request origin for matching our whitelists.
    First tries dev whitelists (that list is expected to hold
    either complete URLs or mere protocols, e.g. "chrome-extension://").
    Then the tld whitelist is tried, which is expected to hold
    only the top-level domains.
    Returns True if origin is allowed, False otherwise.
    """
    if not origin:
        return True

    settings = get_settings()
    dev_whitelist = parse_csv_list(settings.get('spynl.dev_origin_whitelist',
                                                ''))
    dev_list_urls = [url for url in dev_whitelist if not url.endswith('://')]
    origin_allowed = origin in dev_list_urls
    dev_list_protocols = [url for url in dev_whitelist if url.endswith('://')]
    for protocol in dev_list_protocols:
        if origin.startswith(protocol):
            origin_allowed = True
    if not origin_allowed:
        try:
            tld = get_tld(origin)
        except (TldBadUrl, TldDomainNotFound):
            tld = origin  # dev domains like e.g. 0.0.0.0:9000 will fall here
        tld_whitelist = parse_csv_list(settings.get(
            'spynl.tld_origin_whitelist', ''))
        if tld in tld_whitelist:
            origin_allowed = True
    return origin_allowed


def get_header_args(request):
    """Return a dictionary with arguments passed as headers."""
    # these require a spynl-specific prefix to be recognized
    headers = {key: value for key, value in request.headers.items()
               if key.lower().startswith('x-spynl-')}
    # We might also get the session id and client IP address with the headers
    for key in request.headers.keys():
        if key.lower() == 'sid':
            headers['sid'] = request.headers[key]
        if key == 'X-Forwarded-For':
            headers['X-Forwarded-For'] = request.headers[key]

    return headers


def get_parsed_body(request):
    """Return the body of the request parsed if request was POST or PUT."""
    settings = get_settings()
    body_parser = settings.get('spynl.post_parser')

    if request.method in ('POST', 'PUT'):
        if body_parser:
            request.parsed_body = body_parser(request)
        else:
            request.parsed_body = {} if not request.body \
                                     else json.loads(request.body)
    else:
        # disregard any body content if not a POST of PUT request
        request.parsed_body = {}

    return request.parsed_body


def unify_args(request):
    """
    Make one giant args dictonary from GET, POST, headers and cookies and
    return it. On the way, create r.parsed_body and r.parsed_get as well.

    It is possible to provide a custom parser for the POST body in the
    settings. Complex data can be given via GET as a JSON string.
    GET would overwrite POST when parameter names collide.
    """
    args = {}
    # get headers first, they might be useful for parsing the body
    args.update(get_header_args(request))
    # get POST data
    args.update(get_parsed_body(request))
    # get GET args, can be written in JSON style
    # args.update(urlson.loads_dict(request.GET))
    # TODO: needs some refactoring - maybe urlson can actually do this parsing
    # for us. We don't know the context yet.
    from spynl.main.serial import objects
    context = hasattr(request, 'context') and request.context or None
    args.update(json.loads(json.dumps(urlson.loads_dict(request.GET)),
                           object_hook=objects.SpynlDecoder(context=context)))

    request.endpoint_method = find_view_name(request)

    # get cookies, but do not overwrite explicitly given settings
    for key in request.cookies:
        if key not in args:
            args[key] = request.cookies[key]

    # we actually want the sid to live as a header from here on out.
    # It can come in other ways as well (e.g. in GET) for convenience,
    # but we agree for it to live in one place.
    if args.get('sid'):
        request.headers['sid'] = args['sid']
        del args['sid']

    return args


def find_view_name(request):
    """find the view name
    TODO: I believe this is not completely generic.
    """
    name = None

    if request.matchdict and 'method' in request.matchdict:  # a route was matched
        name = request.matchdict['method']
    else:
        name = request.path_info
    if name.startswith('/'):
        name = name[1:]

    if hasattr(request, 'matched_route')\
      and request.matched_route:
        if name in request.matched_route.name:
            # method  name was not in the URL
            if request.method == 'POST':
                name = 'edit'
            elif request.method == 'GET':
                name = 'get'

    return name


def get_user_info(request, purpose=None):
    """
    Spynl.main has no user model. This function allows the use of a
    user_info function defined in a plugin, by setting it to the
    'user_info' setting in the plugger.py of the plugin. If no
    other function is defined, it uses _user_info instead.
    The user_info function should return a dictionary with
    information about the (authenticated) user. If no information is
    available it should return an empty dictionary.
    """
    if request.registry.settings.get('user_info_function') is not None:
        return request.registry.settings['user_info_function'](request, purpose)
    return _get_user_info(request)


def _get_user_info(request):
    """
    Function to get user information as a dictionary. In spynl.main the
    only user information we can get is the ip address.
    """
    ipaddress = get_user_ip(request)
    return dict(ipaddress=ipaddress)


def get_user_ip(request):
    """ Get the ipaddress of the user """
    ipaddress = request.environ.get('REMOTE_ADDR', None)
    # Load balancers overwrite ipaddress,
    # so we prefer the forward header EBS sets
    if 'X-Forwarded-For' in request.headers.keys():
        ipaddress = request.headers['X-Forwarded-For']
    return ipaddress


def get_err_source(original_traceback=None):
    """Use this when an error is handled to get info on where it occured"""
    try:  # carefully try to get the actual place where the error happened
        if not original_traceback:
            original_traceback = sys.exc_info()[2]  # class, exc, traceback
        first_call = traceback.extract_tb(original_traceback)[-1]
        return dict(module=first_call[0],
                    linenr=first_call[1],
                    method=first_call[2],
                    src_code=first_call[3])
    except Exception:
        return 'I was unable to retrieve error source information.'


def renderer_factory(info):
    """
    Normally responses are rendered as bare JSON, but this factory will look
    into the settings for other requested renderers first.
    """
    if hasattr(info, 'settings'):
        settings = info.settings
    if settings and 'spynl.renderer' in settings:
        return settings['spynl.renderer']
    return json_renderer_factory(None)


def is_package_installed(package_name):
    """
    Return True if the package has been pip-installed
    in the current environment
    """
    # importing pip here so pip>=1.5.1 does not cause cyclic package troubles
    # when starting the spynl application
    import pip
    return package_name in [i.key for i in pip.get_installed_distributions()]



def get_logger(name=None):
    """Return the Logger object with the given name."""
    if not name:
        name = __name__

    return logging.getLogger(name)


def parse_value(value, class_info):
    '''
    Parse a value. class_info is expected to be a class or a list
    of classes to try in order.
    Raises SpynlException exception if no parsing was possible.
    '''
    if isclass(class_info):
        try:
            return class_info(value)
        except:
            raise SpynlException(_(
                'parse-value-exception-as-class',
                default='"${value}" could not be parsed as ${class}',
                mapping={'value': value, 'class': class_info.__name__}))

    if hasattr(class_info, '__iter__'):
        for cl in class_info:
            if not isclass(cl):
                raise SpynlException(_(
                    'parse-value-exception-not-class',
                    default='${class} is not a class. Cannot parse '
                            'value "${value}".',
                    mapping={'class': cl, 'value': value}))
            try:
                return cl(value)
            except Exception:
                pass
    raise SpynlException(_(
        'parse-value-exception-any-class',
        default='"${value}" could not be parsed into any class in ${classes}',
        mapping={'value': value,
                 'classes': [cl.__name__ for cl in class_info]}))



def parse_csv_list(csv_list):
    """Parse a list of CSV values."""
    return [i.strip() for i in csv_list.split(',')]


def get_yaml_from_docstring(doc_str, load_yaml=True):
    """
    Load the YAML part (after "---") from the docstring of a Spynl view.

    if load_yaml is True, return the result of yaml.load, otherwise return
    as string.
    """
    if doc_str:
        yaml_sep = doc_str.find('---')
    else:
        yaml_sep = -1

    if yaml_sep != -1:
        yaml_str = doc_str[yaml_sep:]
        if load_yaml:
            return yaml.load(yaml_str)
        else:
            return yaml_str
    return None


def required_args(*arguments):
    """Call the decorator that checks if required args passed in request."""
    def outer_wrapper(func):
        """Return the decorator."""
        @wraps(func)
        def inner_wrapper(*args):
            """
            Raise if a required argument is missing or is empty.

            Decorator checks if request.args were the expected <*arguments> of
            the current endpoint.
            """
            request = args[-1]  # request is always the last argument
            for required_arg in arguments:
                if request.args.get(required_arg, None) is None:
                    raise MissingParameter(required_arg)
            if len(getargspec(func).args) == 1:
                return func(request)
            else:
                return func(*args)
        return inner_wrapper
    return outer_wrapper


def send_exception_to_external_monitoring(user_info=None, exc_info=None,
                                          metadata=None, endpoint=None):
    """Send exception info to online services for better monitoring
    The user_info param can be added so the services can display which user
    was involved.
    The exc_info parameter should only be passed in if a
    different exception than the current one on the stack should be sent.
    The metadata parameter can be used for any extra information.
    The endpoint parameter is sent under the tags Sentry parameter so
    exceptions can be filtered in their website by endpoint.
    """
    log = get_logger()
    settings = get_settings()
    if exc_info is None:
        exc_info = sys.exc_info()
    # send Exception to Sentry if the raven package is installed
    # and sentry is configured
    try:
        if is_package_installed('raven') and settings.get('spynl.sentry_key')\
          and settings.get('spynl.sentry_project'):
            import raven  # pylint: disable=import-error
            spynl_env = settings.get('spynl.spynl_environment', 'dev')
            spynl_function = settings.get('spynl.function', 'all')
            client = raven.Client(
                dsn='https://{}@app.getsentry.com/{}'.format(
                    settings.get('spynl.sentry_key'),
                    settings.get('spynl.sentry_project')),
                release=spynl_version,
                site='Spynl [{}-{}]'.format(spynl_env, spynl_function),
                processors=('raven.processors.SanitizePasswordsProcessor',))
            if user_info is not None:
                client.user_context(user_info)
            data = dict(exc_info=exc_info, extra=metadata)
            if endpoint is not None:
                data.update(tags=dict(endpoint=endpoint))
            client.captureException(**data)
    except Exception as e:
        log.warning('Could not send Exception to Sentry: %s', e)

    # tell NewRelic about user information if the newrelic package is installed
    # (the rest of the configuration of NewRelic is ini-file-based)
    try:
        if is_package_installed('newrelic') and user_info is not None:
            import newrelic.agent  # pylint: disable=import-error
            for key in user_info:
                # do not include ipaddress for privacy
                if key != 'ipaddress':
                    newrelic.agent.add_custom_parameter(key, user_info[key])

    except Exception as e:
        log.warning('Could not add user info to NewRelic on exception: %s', e)


def log_error(exc, request, top_msg, error_type, error_msg):
    """
    Log the error from an error view to the log, and to external monitoring.
    Make sure the __cause__ of the exception is used.
    """
    log = get_logger()

    user_info = get_user_info(request, purpose='error_view')

    exc_info = sys.exc_info()
    if hasattr(exc, '__cause__') and exc.__cause__ is not None:
        exc_info = (exc.__cause__.__class__, exc.__cause__, sys.exc_info()[2])
    metadata = dict(user=user_info,
                    url=request.path_url,
                    err_source=get_err_source(sys.exc_info()[2]),
                    detail=getattr(exc, 'detail', None))
    log.error(top_msg, error_type, str(error_msg),
              exc_info=exc_info,
              extra=dict(meta=metadata,
                         payload=str(request.body)[:500]))

    send_exception_to_external_monitoring(user_info=user_info,
                                          exc_info=exc_info,
                                          metadata=metadata)


def is_production_environment(or_test=False):
    """Return if environment is either beta/production or not."""
    settings = get_settings()
    allowed_envs = ['beta', 'production']
    if or_test:
        allowed_envs.append('test')
    return settings.get('spynl.spynl_environment') in allowed_envs



import os
import contextlib
from invoke.exceptions import Exit

from spynl.main.pkg_utils import get_spynl_package, get_spynl_packages


def resolve_repo_names(repos_param, complain_not_installed=True):
    """resolve repo string parameter into a list of repos"""
    installed_repo_names = [r.project_name for r in get_spynl_packages()]
    installed_repo_names.sort()
    if repos_param != '_all':
        given_names = repos_param.split(',')
        for name in given_names:
            if name not in installed_repo_names and complain_not_installed:
                raise Exit("Repo %s is not installed. Exiting ..." % name)
        return given_names
    return installed_repo_names


@contextlib.contextmanager
def repo(reponame=None):
    """Change to this repo directory during this context"""
    curdir = os.getcwd()
    try:
        if reponame is not None:
            repo = get_spynl_package(reponame)
            if repo is not None:
                os.chdir(repo.location)
        yield
    finally:
        os.chdir(curdir)


@contextlib.contextmanager
def chdir(dirname=None):
    """Change to this directory during this context"""
    curdir = os.getcwd()
    try:
        if dirname is not None:
            os.chdir(dirname)
        yield
    finally:
        os.chdir(curdir)


def assert_response_code(response, exp_code):
    """
    Make it easy to expect one or more HTTP status codes
    in a requests library response.
    """
    code = response.status_code
    if isinstance(exp_code, tuple):
        if not code in exp_code:
            raise Exit("Code %s is not in %s when testing %s"
                       % (code, exp_code, response.request.url))
    else:
        if not code == exp_code:
            raise Exit("Code %s was expected to be %s when testing %s "
                       % (code, exp_code, response.request.url))



"""
Set-up for ini documentation list of dicts

In the ini_doc list, only use 'yes' or 'no' for the required field,
as this is used in the code to check if all required settings are set.
"""

from spynl.main.exceptions import SpynlException


ini_description = '''
These are the settings that can be added to the ini file under the heading
[app:main]. This table only lists settings specific to Spynl. Pyramid settings
and Pyramid Mail configuration should also be put in [app:main], but are not
listed here.  Spynl will not start if one of the required settings is not
present in the ini file. <br>
The pyramid asbool function returns the boolean value True if the case-lowered
value of string input s is any of t, true, y, on, or 1, otherwise return the
boolean value False. If s is the value None, return False. If s is already one
of the boolean values True or False, return it. <br>
N.B. It is not possible to add comments after the setting,
comments should be on their own line.
'''

ini_doc = [{'name': 'spynl.date_systemtz',
            'plugin': '',
            'required': 'no',
            'default': 'UTC',
            'info': 'Timezone that is used if no time zone is explicitly '
                    'given and also no time zone is found in or to be '
                    'used from the user object.'},

           {'name': 'spynl.date_format',
            'plugin': '',
            'required': 'no',
            'default': '%Y-%m-%dT%H:%M:%S%z',
            'info': 'Format for dates.'},

           {'name': 'spynl.languages',
            'plugin': '',
            'required': 'no',
            'default': 'en',
            'info': 'A csv string describing the languages supported by the '
                    'application. Spynl has no dialect support, so only '
                    'two-letter language codes should be provided. '
                    'The first language is considered to be the preferred '
                    'language.'},

           {'name': 'spynl.domain',
            'plugin': '',
            'required': 'no',
            'default': 'localhost',
            'info': 'Used for setting the cookie-domain, and for a default '
                    'no-reply emailaddress.'},

           {'name': 'spynl.languages',
            'plugin': '',
            'required': 'no',
            'default': 'nl',
            'info': 'Languages (e.g. de,nl) which Spynl should translate '
                    'to (other than english). This affects how the task '
                    '<spynl dev.translations> works. Use a comma-separated '
                    ' list for more than one.'},
 
           {'name': 'spynl.tld_origin_whitelist',
            'plugin': '',
            'required': 'no',
            'default': '',
            'info': 'Comma-separated whitelist for allowed origins. '
                    'It is expected to hold only the top-level domains, '
                    ' e.g. "google.com".'},

           {'name': 'spynl.dev_origin_whitelist',
            'plugin': '',
            'required': 'no',
            'default': '',
            'info': 'Comma separated whitelist for allowed origins which is '
                    'usable for a development context. It is expected to hold'
                    ' either complete domains or mere protocols, e.g. '
                    '"chrome-extension://"..'},

           {'name': 'spynl.pretty',
            'plugin': '',
            'required': 'no',
            'default': 'false',
            'info': 'Pretty printing for development. Is read with Pyramid '
                    'asbool function.'},

           {'name': 'spynl.session',
            'plugin': '',
            'required': 'no',
            'default': 'Pyramid',
            'info': 'Sets the name of the session class to use. Spynl will '
                    'look this class up under settings[<name>]. '
                    'The class should conform to the interface '
                    'pyramid.interfaces.ISession. The value "Pyramid" or '
                    'omitting this setting will result in the use of the '
                    'Session class from beaker.session'},
 
           {'name': 'spynl.ops.scm_type',
            'plugin': '',
            'required': 'no',
            'default': 'git',
            'info': 'The source control management system used by your Spynl '
                    'plugin repositories use. One of "git" or "hg", defaults '
                    ' to "git".'},
 
           {'name': 'spynl.ops.scm_team',
            'plugin': '',
            'required': 'no',
            'default': '',
            'info': 'The source control management provider+team where '
                    'your Spynl plugin repositories can be found, e.g. '
                    '"github.com/myteam" or "bitbucket.org/myteam".'},
           
           {'name': 'spynl.ops.environment',
            'plugin': '',
            'required': 'no',
            'default': '',
            'info': 'The environment Spynl is running in. This is quite '
                    'useful for things like error log aggregation.'},

           {'name': 'spynl.ops.function',
            'plugin': '',
            'required': 'no',
            'default': '',
            'info': 'The function this Spynl node is performing. This '
                    'can be useful for error log aggregation if you '
                    ' employ more than one Spynl applications.'},

           {'name': 'spynl.ops.jenkins_url',
            'plugin': '',
            'required': 'no',
            'default': '',
            'info': 'The URL of your Jenkins server.'},

           {'name': 'spynl.ops.dev_url',
            'plugin': '',
            'required': 'no',
            'default': '',
            'info': 'The URL where your development app can be '
                    'accessed. Needed to run smoke tests after deploys.'},

           {'name': 'spynl.ops.dev_domain',
            'plugin': '',
            'required': 'no',
            'default': '',
            'info': 'The top level domain on which your development app '
                    'resides. This is used to get the cookie domain '
                    'correct.'},

           {'name': 'spynl.ops.ecr.dev_tasks',
            'plugin': '',
            'required': 'no',
            'default': 'dev',
            'info': 'The tasks supported in your development AWS ECR.'},

           {'name': 'spynl.ops.ecr.dev_url',
            'plugin': '',
            'required': 'no',
            'default': '',
            'info': 'The URL of your AWS ECR used for development. Include '
                    'AWS profile name like this: profile@ecr-url.'},

           {'name': 'spynl.ops.ecr.prod_url',
            'plugin': '',
            'required': 'no',
            'default': '',
            'info': 'The URL of your AWS ECR used for production. Include '
                    'AWS profile name like this: profile@ecr-url.'},

           {'name': 'spynl.sentry.key',
            'plugin': '',
            'required': 'no',
            'default': '',
            'info': 'The key used to connect to Sentry. Leave empty if '
                    'you are not using Sentry.'},

           {'name': 'spynl.sentry.project',
            'plugin': '',
            'required': 'no',
            'default': '',
            'info': 'The project identifier Spynl reports to in Sentry.'},

           {'name': 'spynl.newrelic.key',
            'plugin': '',
            'required': 'no',
            'default': '',
            'info': 'The key used to connect to NewRelic. Leave empty if '
                    'you are not usig NewRelic.'}

          ]
def get_ini_doc_setting(name):
    """
    Returns the documentation entry of a specific setting or None

    Because ini_doc is a list, this function is needed to get the documentation
    of a specific setting.
    """
    for setting in ini_doc:
        if setting['name'] == name:
            return setting

    return None


def check_required_settings(config):
    """check whether all required ini settings are set"""
    for setting in ini_doc:
        if (setting['required'] == 'yes' and
                setting['name'] not in config.registry.settings):
            msg = 'Please set {} in the Pyramid ini file'.format(
                setting['name'])
            raise SpynlException(msg)
        # setting['required'] can only be yes or no (no accidental true's)
        if setting['required'] not in ('no', 'yes'):
            msg = ('The value for required in ini_doc can only be yes or no, '
                   'please change this for setting: {}'
                   .format(setting['name']))
            raise SpynlException(msg)





"""Handle XML content"""

from xml.etree.ElementTree import fromstring, ParseError
from xml.sax import saxutils
import re

from spynl.main.serial import objects
from spynl.main.serial.exceptions import MalformedRequestException


expression = re.compile(r'^\s*\<')


def loads(body, headers=None, context=None):
    """return body as XML"""
    try:
        root = fromstring(body)
    except ParseError as e:
        # pylint: disable=E1101
        raise MalformedRequestException('application/xml', str(e))

    dic = __loads(root, True)
    return objects.SpynlDecoder(context=context)(dic)


def __loads(element, force_dict=False):
    """Recurse through etree node, return parsed structure"""
    if len(element) and element.get('type') != 'collection' or force_dict:
        # this is a mapping (dict)
        result = {}
        for field in element:
            result[field.tag] = __loads(field)
    elif element.get('type') == 'collection':
        # this is a collection (list)
        result = [__loads(item) for item in element]
    else:
        # this is a value
        result = None
        if element.text:
            result = element.text.strip()

    return result


def dumps(body, pretty=True):
    """return XML body as string"""
    result = __dumps(body)

    if pretty:
        result = prettify_xml(result)

    # concatenating bytes and str objects to one str
    ustr = ''
    for item in result:
        ustr += str(item)

    return '<response>{}{}</response>'.format(pretty * '\n', ustr)


def __dumps(value):
    """Recurse through dict/list structure, returning XML text"""
    result = []

    if isinstance(value, (list, tuple, set)):
        for item in value:
            result.append('<item>')
            result.extend(__dumps(item))
            result.append('</item>')
    elif isinstance(value, dict):
        for field, value in value.items():
            if field.startswith('$'):  # e.g. query operators
                field = field[1:]

            alpha = re.match(r'\D', field)  # alphanumeric field name?

            start_tag = '<' + (field if alpha else 'item')
            if not alpha:
                start_tag += ' key="{}"'.format(field)
            if isinstance(value, (list, tuple)):
                start_tag += ' type="collection"'
            start_tag += '>'
            end_tag = '</{}>' if alpha else '</item>'

            result.append(start_tag.format(field))
            result.extend(__dumps(value))
            result.append(end_tag.format(field))
    else:
        if isinstance(value, str):
            value = saxutils.escape(value)
        result.append(objects.encode(value))

    return result


def sniff(body):
    """sniff to see if body is xml"""
    return bool(re.match(expression, body))


def prettify_xml(xmlstr):
    """add indentation to create pretty xml"""
    indent = 1
    for i in range(len(xmlstr)):
        item = xmlstr[i]
        if item.startswith('</'):
            indent -= 1
            tag = xmlstr[i - 1].startswith('<')
            yield tag * ' ' * indent * 4 + item + '\n'
        elif item.startswith('<'):
            tag = xmlstr[i + 1].startswith('<')
            yield ' ' * indent * 4 + item + tag * '\n'
            indent += 1
        else:
            yield item
    raise StopIteration()



"""Command-line tool to test the (de)serialisation live."""

from sys import stdin, stdout
from argparse import ArgumentParser

from spynl.main.serial import negotiate_content_type, loads, dumps


def main():
    """main function for converting between formats"""
    parser = ArgumentParser(description='Convert between data formats.')
    parser.add_argument('--output', metavar='FILENAME',
                        dest='output_filename', default=None,
                        help='filename where output is stored,'
                             ' STDOUT by default')
    parser.add_argument('--output-type', metavar='TYPE',
                        dest='output_type', default='json',
                        help='output type e.g. JSON or XML etc.')
    parser.add_argument('--input-type', metavar='TYPE',
                        dest='input_type', default=None,
                        help='suggested input type. See --output-type')
    args = parser.parse_args()

    request = stdin.read()

    request = loads(request, negotiate_content_type(request, args.input_type))
    response = dumps(request, negotiate_content_type('', args.output_type))

    output = open(args.output_filename, 'w')\
        if args.output_filename else stdout
    output.write(response + '\n')

    if args.output_filename:
        output.close()



from invoke import Program, Collection

from spynl.main.version import __version__ as spynl_version
from cli.dev import tasks as dev_tasks
from cli.dev import hg
from cli.ops import tasks as ops_tasks


ns = Collection()
ns.add_collection(dev_tasks, 'dev')
ns.add_collection(hg)
ns.add_collection(ops_tasks, 'ops')
program = Program(version=spynl_version, namespace=ns)



"""
Functions for handling dates and times.
"""

from datetime import datetime
import dateutil.parser  # pylint: disable=E0611
from pytz import utc, timezone

from spynl.main.utils import get_request, get_settings, get_user_info


def now(tz=None):
    """Current time, with timezone localised."""
    return localize_date(datetime.utcnow(), tz=tz)


def date_format_str():
    """Get the date format from the .ini file, or set default."""
    return get_settings().get('spynl.date_format', '%Y-%m-%dT%H:%M:%S%z')


def date_to_str(when):
    """Convert date to string according to settings format"""
    return when.strftime(date_format_str())


def date_from_str(dstr):
    """"
    Parses a string to a date/time and adds current timezone if no timezone
    information is present.
    """
    # slightly slower than datetime.strptime
    # but can handle different inputs
    _when = dateutil.parser.parse(dstr)  # pylint: disable=no-member
    if not _when.tzinfo:  # pylint: disable=E1103
        _when = utc.localize(_when)
    return _when


def localize_date(when, user_specific=True, tz=None):
    """
    Localises datetime objects to a time zone.
    If no time zone is explicitly given and also no time zone is found
    in or to be used from the user information, then system timezone is used.
    If 'when', the datetime object, has itself has no timezone info,
    then we assume it represents UTC time.
    """
    if not tz:
        tz = get_settings().get('spynl.date_systemtz', 'UTC')
        if user_specific:
            request = get_request()
            if request:  # e.g. in testing there might be no request
                user_info = get_user_info(request)
                if user_info.get('tz'):
                    tz = user_info.get('tz')
    if not when.tzinfo:
        when = when.replace(tzinfo=utc)
    return when.astimezone(timezone(tz))
